% Simulation code for V2X
% by Suhua Tang @ UEC
clear, close all
[propParam, macParam, protoParam] = get_BasicSimParam;

simParam.simTime = [0 12000];  % 12000 ms
simParam.simSeed = 1;
simParam.maxSpeed= 20;         % 20 m/s, 72 km/h
simParam.kFactor = 0;

simParam.txRng = 201;          % 201 m
simParam.minRRI = 5;
simParam.maxRRI = 15;
simParam.rxRSSITh = -80;       % dBm, min RSSI for correct reception
simParam.rxSINRth = 5;         % dB,  min SINR for correct reception

simParam.resultPath = '..\result_V2X';
simParam.extraName = sprintf('K%d', simParam.kFactor);

simParam.simProto= protoParam.PROTO_V2X_SPS;
simParam.log_file= sprintf('%s\\Log-%s-S%d-%s.txt', simParam.resultPath, ...
				protoParam.PROTO_NAME{simParam.simProto}, simParam.simSeed, simParam.extraName);


%%%%%%%%%%%%%%%%%%%% node placement and fading parameters %%%%%%%%%%%%%%%%%%%%
nodes = get_NodePlacement;
numOfNodes  = length(nodes);
simParam.fc_lambda = propParam.fc_lambda;
fadingParam = get_ChannelFadingParameter(numOfNodes, simParam);

for nId=1:numel(nodes)
	%%%%% flow parameters
	nodes(nId).nodeId = nId;
	nodes(nId).txInterval = 20;  % interval = 100 ms
	nodes(nId).packetSize= 128;  % packet size
    nodes(nId).reselCnt = randi([simParam.minRRI, simParam.maxRRI]);  % initialize reselection counter
    nodes(nId).P = 1.0;  % initialize reselection probability
    nodes(nId).selWin = [1, nodes(nId).txInterval];
    nodes(nId).txRes = randi(nodes(nId).selWin);  % selected resource
    nodes(nId).txResReserv = 0;                   % reserved resource for next TX
    nodes(nId).totalPktNum = (simParam.simTime(2)-simParam.simTime(1)) / nodes(nId).txInterval;
    nodes(nId).currPktId = 1;
    nodes(nId).statRxStat = zeros(nodes(nId).totalPktNum,3);
    nodes(nId).rxHist = zeros(1000,5);   % record time, RSSI, TX-Id, TX-resource, TX-resource reservation
end


rand('state', simParam.simSeed);
fprintf(2, '************************** %s=protocol, Seed=%d, %s ***************************\n', ...
	protoParam.PROTO_NAME{simParam.simProto}, simParam.simSeed, simParam.extraName);


%%%%%%%%%%%%%%%%%%%% actual simulation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fid = fopen(simParam.log_file, 'w');
for currTime = simParam.simTime(1):1:simParam.simTime(2)
    % determine nodes that transmit
    txResSet = [nodes(:).txRes];
    indicator = (txResSet == currTime);
    txNodeSet = [nodes(indicator).nodeId];

    % if currTime == 5600
    %     fprintf('currTime = %d\n', currTime);
    % end

    if isempty(txNodeSet)
        for idx = 1:numel(nodes)  % shift RX history
            nodes(idx).rxHist = [ [currTime,-200,0,0,0]
                nodes(idx).rxHist(1:end-1,:)];    
        end
        continue
    end

    % update resource reservation info
    for idx = 1:numel(txNodeSet)
        nId = txNodeSet(idx);
        if (nodes(nId).reselCnt > 0)
            nodes(nId).txResReserv = nodes(nId).txRes + nodes(nId).txInterval;
        else
            nodes(nId).txResReserv = 0; % not reserve resource
        end
    end

    % get instantaneous position of all nodes
    position = get_NodeCurrPosition(nodes, (1:numOfNodes), currTime/1000);

    % compute RSSI
    distTx2Rx = zeros(numel(nodes), numel(txNodeSet));
    rssiMatDb = zeros(numel(nodes), numel(txNodeSet));
    for idx = 1:numel(txNodeSet)
        nId = txNodeSet(idx);
		txPos = position(nId);
		distTx2Rx(:,idx)   = abs(txPos-position);
        distTx2Rx(nId,idx) = simParam.txRng * 2;
		SNR_dB     = get_ChannelSNRdB(distTx2Rx(:,idx), propParam);
		% nodePairSet= [repmat(nId, numel(nodes), 1), (1:numel(nodes))'];
		% fading_dB  = get_ChannelFadingdB(currTime/1000, nodePairSet, fadingParam);
		% RSSI_dB    = SNR_dB + fading_dB + propParam.NoisePower_dBm;
        RSSI_dB    = SNR_dB + propParam.NoisePower_dBm;
        RSSI_dB(nId) = -200;
        rssiMatDb(:,idx) = RSSI_dB;
    end

    % compute packet rx prob
    if (numel(txNodeSet)==1) % only one nodes TX, no collision
        nId = txNodeSet;
        inRng = distTx2Rx <= simParam.txRng;
        isCorrect = rssiMatDb >= simParam.rxRSSITh;
        nodes(nId).statRxStat(nodes(nId).currPktId,:) = [sum(isCorrect & inRng), sum(inRng), currTime];
        nodes(nId).currPktId = nodes(nId).currPktId + 1;
        % record time, RSSI, TX-Id, TX-resource, TX-resource reservation
        for idx=1:numel(nodes)
            if nodes(idx).nodeId ~= nId % update RX history of RX node
                if isCorrect(idx) && inRng(idx)
                    nodes(idx).rxHist = [ [currTime,rssiMatDb(idx),nId,nodes(nId).txRes,nodes(nId).txResReserv]
                        nodes(idx).rxHist(1:end-1,:)];
                else
                    nodes(idx).rxHist = [ [currTime,rssiMatDb(idx),0,-1,-1]    % weak signal, not correctly received
                        nodes(idx).rxHist(1:end-1,:)];    
                end
            else % TX node, shift RX history
                nodes(idx).rxHist = [ [currTime,-200,idx,0,0]
                    nodes(idx).rxHist(1:end-1,:)];                
            end
        end

    else % more than 2 nodes TX, a collision occurs
        rxCorrectMat = false(numel(nodes), numel(txNodeSet));
        for rId = 1:numel(nodes)
            [~, index] = sort(rssiMatDb(rId,:),'descend');
            if rssiMatDb(rId,index(1)) >= simParam.rxRSSITh ...
                    && rssiMatDb(rId,index(1)) - rssiMatDb(rId,index(2)) >= simParam.rxSINRth
                % capture
                rxCorrectMat(rId,index(1)) = true;
            end
            % record time, RSSI, TX-Id, TX-resource, TX-resource reservation
            if ~ismember(rId,txNodeSet) % update RX history of RX node
                txId = txNodeSet(index(1));
                if rxCorrectMat(rId,index(1))
                    nodes(rId).rxHist = [ [currTime,rssiMatDb(rId,index(1)),txId,nodes(txId).txRes,nodes(txId).txResReserv]  % correctly received
                        nodes(rId).rxHist(1:end-1,:)];
                else
                    nodes(rId).rxHist = [ [currTime,rssiMatDb(rId,index(1)),-1,-1,-1]    % collision, not correctly received
                        nodes(rId).rxHist(1:end-1,:)];                    
                end
            else % TX node, shift RX history
                nodes(rId).rxHist = [ [currTime,-200,rId,0,0]
                    nodes(rId).rxHist(1:end-1,:)];                    
            end
        end
        for idx = 1:numel(txNodeSet)
            nId = txNodeSet(idx);
            inRng = distTx2Rx(:,idx) < simParam.txRng;
            isCorrect = rxCorrectMat(:,idx);
            nodes(nId).statRxStat(nodes(nId).currPktId,:) = [sum(isCorrect & inRng), sum(inRng), currTime];
            nodes(nId).currPktId = nodes(nId).currPktId + 1;
        end
    end

    % update reselCnt of nodes
    for idx = 1:numel(txNodeSet)
        nId = txNodeSet(idx);
        if (nodes(nId).reselCnt == 0)
            nodes(nId).reselCnt = randi([simParam.minRRI, simParam.maxRRI]);
            nodes(nId).selWin = currTime + [1, nodes(nId).txInterval];
            % select resource randomly
            % nodes(nId).txRes = randi(nodes(nId).selWin);
            % select resource by SPS
            nodes(nId).txRes = SPS_RA(nId,nodes(nId).selWin,nodes(nId).rxHist);
        else
            % decrease reselection counter & set next tx timing
            nodes(nId).reselCnt = nodes(nId).reselCnt - 1;
            nodes(nId).txRes = nodes(nId).txRes + nodes(nId).txInterval;
        end
    end

    % nId = 5;
    % if ismember(nId, txNodeSet)
    %     fprintf('currTime = %5d, %2d, %2d, %5d\n', currTime, nId, nodes(nId).reselCnt, nodes(nId).txRes);
    % end
end

% printf packet receive rate
nId = 5;
stat = nodes(nId).statRxStat(1:nodes(nId).currPktId-1,:);
fprintf('node = %d, PRR = %.3f\n', nId, mean(stat(:,1)./stat(:,2)));

nId2= 10;
collisionSlot = intersect(nodes(nId).statRxStat(1:nodes(nId).currPktId-1,3), ...
                          nodes(nId2).statRxStat(1:nodes(nId2).currPktId-1,3));

x1 = mod(nodes(nId).statRxStat(1:nodes(nId).currPktId-1,3),  nodes(nId).txInterval);
index = x1==0;
x1(index) = nodes(nId).txInterval;
x2 = mod(nodes(nId2).statRxStat(1:nodes(nId2).currPktId-1,3),nodes(nId2).txInterval);
index = x2==0;
x2(index) = nodes(nId2).txInterval;

figure(1)
plot(x1, 'g+');
hold on;
plot(x2, 'bx')
xlabel('Index of frames');
ylabel('Index of resource');
axis([1, 200, 1, 20]); % axis([1, numel(x1), 1, 20]);
legend('No.5', 'No.10');

return;


